#!/usr/bin/env python
# -*- coding: utf-8 -*-
import argparse
from argparse import ArgumentParser
from pathlib import Path
from time import time

import pydicom

from ..dicom import decompress


def get_parser(parser: ArgumentParser = ArgumentParser()) -> ArgumentParser:
    parser.add_argument("path", type=Path, help="DICOM file to decompress")
    parser.add_argument("dest", type=Path, help="destination DICOM filepath")
    parser.add_argument(
        "-s", "--strict", default=False, action="store_true", help="if true, raise an error on uncompressed input"
    )
    parser.add_argument(
        "-g", "--gpu", default=False, action="store_true", help="use NVJPEG2K accelerated decompression"
    )
    parser.add_argument(
        "-b", "--batch-size", default=None, type=int, help="batch size for NVJPEG2K accelerated decompression"
    )
    parser.add_argument("-v", "--verbose", default=False, action="store_true", help="print NVJPEG2K outputs")
    parser.add_argument(
        "-t", "--test", default=False, action="store_true", help="test equivalence to CPU decompression"
    )
    return parser


def main(args: argparse.Namespace) -> None:
    path = Path(args.path)
    dest = Path(args.dest)
    if not path.is_file():
        raise FileNotFoundError(path)
    if not dest.parent.is_dir():
        raise NotADirectoryError(dest.parent)

    start_time = time()
    with pydicom.dcmread(path) as dcm:
        dcmread_time = time() - start_time
        dcm = decompress(dcm, strict=args.strict, use_nvjpeg=args.gpu, batch_size=args.batch_size, verbose=args.verbose)
        assert not dcm.file_meta.TransferSyntaxUID.is_compressed
        dcm.save_as(dest)
    end_time = time()
    if args.verbose:
        total_time = end_time - start_time
        print(f"pydicom.dcmread time (s): {dcmread_time}")
        print(f"Total time (s): {total_time}")

    if args.test:
        with pydicom.dcmread(path) as dcm2:
            px1 = dcm.pixel_array
            px2 = dcm2.pixel_array
            assert (px1 == px2).all()


def entrypoint():
    parser = get_parser()
    main(parser.parse_args())


if __name__ == "__main__":
    entrypoint()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import runpy
import sys
from pathlib import Path

import numpy as np
import pydicom
import pytest
from pydicom.data import get_testdata_file
from pydicom.uid import ExplicitVRLittleEndian, RLELossless

from dicom_utils.dicom_factory import DicomFactory


@pytest.fixture
def get_dicom_file(tmp_path):
    def func(compressed: bool = True):
        get_testdata_file("CT_small.dcm")
        dest = Path(tmp_path, "file.dcm")
        fact = DicomFactory(
            NumberOfFrames=32,
            Rows=128,
            Columns=128,
            TransferSyntaxUID=RLELossless if compressed else ExplicitVRLittleEndian,
        )
        dcm = fact()
        dcm.save_as(dest)
        return dest

    return func


@pytest.mark.parametrize("start_compressed", [True, False])
def test_decompress(get_dicom_file, tmp_path, start_compressed):
    dicom_file = get_dicom_file(start_compressed)
    dest = Path(tmp_path, "output.dcm")
    sys.argv = [
        sys.argv[0],
        str(dicom_file),
        str(dest),
    ]
    runpy.run_module("dicom_utils.cli.project", run_name="__main__", alter_sys=True)

    assert dest.is_file()
    with pydicom.dcmread(dest) as dcm:
        assert not dcm.file_meta.TransferSyntaxUID.is_compressed
        assert isinstance(dcm.pixel_array, np.ndarray)



import re
import secrets
import string
from typing import Callable, Dict, Final, Optional, TypeVar

import pydicom
from dicomanonymizer import anonymize_dataset
from pydicom import Dataset

from .private import MEDCOG_NAME, get_medcog_block, get_medcog_elements, store_medcog_elements
from .tags import Tag


T = TypeVar("T")

FAKE_UID: Final[bytes] = b"0.0.000.000000.0000.00.0000000000000000.00000000000.00000000"
UID_LEN: Final[int] = len(FAKE_UID)
PATIENT_ID_PREFIX: Final[str] = "MEDCOG"
PATIENT_ID_LOOKUP: Final[Dict[str, str]] = {}


def fix_bad_fields(raw_elem, **kwargs):
    try:
        if raw_elem.tag == Tag.NumberOfFrames.tag_tuple and raw_elem.value is None:
            # Value of "None" is non-conformant
            raw_elem = raw_elem._replace(value=b"1", length=1)
        elif raw_elem.tag == Tag.IrradiationEventUID.tag_tuple and len(raw_elem.value) > UID_LEN:
            # The DICOM anonymizer doesn't handle a list of UIDs properly
            raw_elem = raw_elem._replace(value=FAKE_UID, length=UID_LEN)
    except Exception as e:
        # It's possible to throw a `TypeError` if `raw_elem.value == None`, for example.
        # We could check for this and other conditions, but — either way — running this function is not critical.
        print(f"WARNING: `{e}`")

    return raw_elem


pydicom.config.data_element_callback = fix_bad_fields  # type: ignore
pydicom.config.convert_wrong_length_to_UN = True  # type: ignore


def preserve_value(x: T) -> T:
    return x


def str_to_first_int(s: str) -> Optional[int]:
    x = re.findall(r"\d+", s)
    if len(x) > 0:
        return int(x[0])


def anonymize_age(age_str: str) -> str:
    """So few people live into their 90s that an age greater than 89 is considered to be identifying information."""
    age: Optional[int] = str_to_first_int(age_str)
    if age is None:
        return "----"
    elif age > 89:
        return "90Y+"
    else:
        return f"{age:03}Y"


def generate_rand_str(num_chars: int = 32) -> str:
    chars62 = string.ascii_letters + string.digits  # Almost 6-bits
    return "".join(secrets.choice(chars62) for _ in range(num_chars))


def anonymize_patient_id(patient_id: str) -> str:
    if patient_id not in PATIENT_ID_LOOKUP:
        PATIENT_ID_LOOKUP[patient_id] = f"{PATIENT_ID_PREFIX}{generate_rand_str()}"
    return PATIENT_ID_LOOKUP[patient_id]


RuleMap = Dict[Tag, Callable[[T], T]]

DEFAULT_ANON_RULES: Final[RuleMap] = {
    Tag.PatientID: anonymize_patient_id,
    Tag.PatientAge: anonymize_age,
    Tag.PatientSex: preserve_value,
    Tag.CountryOfResidence: preserve_value,
    Tag.EthnicGroup: preserve_value,
    # Institution names should be OK to keep per the following explanation:
    # "Only names of the individuals associated with the corresponding health
    # information (i.e., the subjects of the records) and of their relatives,
    # employers, and household members must be suppressed. There is no explicit
    # requirement to remove the names of providers or workforce members of the
    # covered entity or business associate."
    # https://www.hhs.gov/hipaa/for-professionals/privacy/special-topics/de-identification/index.html#supress
    Tag.InstitutionName: preserve_value,
    # `DerivationDescription` describes operations with which a DICOM image has been modified,
    # and it can be useful for checking settings that may affect image quality.
    Tag.DerivationDescription: preserve_value,
}


def is_anonymized(ds: Dataset) -> bool:
    try:
        get_medcog_block(ds)
        return True
    except KeyError as e:
        assert str(e) == f"\"Private creator '{MEDCOG_NAME}' not found\""
        return False


def anonymize(ds: Dataset, anon_rules: RuleMap = DEFAULT_ANON_RULES) -> None:
    # anonymize_dataset() deletes private elements
    # so we need to store value hashes in the MedCognetics private elements after anonymization
    assert not is_anonymized(ds), "DICOM file is already anonymized"

    overrides = {tag: rule(getattr(ds, tag.name, "")) for tag, rule in anon_rules.items()}
    elements = get_medcog_elements(ds)

    anonymize_dataset(ds, delete_private_tags=True)

    store_medcog_elements(ds, elements)
    for tag, value in overrides.items():
        setattr(ds, tag.name, value)



import os
from pathlib import Path
from typing import Final

import pydicom
import pytest
from pydicom import Dataset

from dicom_utils.cli.strip import strip_pixel_data, to_stripped_dicom


num_dicom_test_files: Final[int] = 3


@pytest.fixture(params=pydicom.data.get_testdata_files("*rgb*.dcm")[:num_dicom_test_files])  # type: ignore
def test_dicom_path(request) -> Dataset:
    return request.param


def test_strip_pixel_data(test_dicom_path) -> None:
    test_dicom = pydicom.dcmread(test_dicom_path)
    assert "PixelData" in test_dicom
    strip_pixel_data(test_dicom)
    assert "PixelData" not in test_dicom


def test_to_stripped_dicom(test_dicom_path, tmpdir) -> None:
    new_dicom_path = Path(tmpdir) / "output.dcm"
    to_stripped_dicom(test_dicom_path, new_dicom_path)
    assert "PixelData" in pydicom.dcmread(test_dicom_path)
    assert "PixelData" not in pydicom.dcmread(new_dicom_path)
    assert os.path.getsize(new_dicom_path) < os.path.getsize(test_dicom_path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import runpy
import shutil
import sys
from pathlib import Path

import pytest


@pytest.fixture
def dicom_file():
    pydicom = pytest.importorskip("pydicom")
    return pydicom.data.get_testdata_file("CT_small.dcm")


@pytest.fixture
def dicom_folder(dicom_file, tmp_path):
    parent = Path(tmp_path)
    parent.mkdir(exist_ok=True)
    file_list = []
    for i in range(3):
        child = Path(dicom_file).with_suffix("").stem
        child = Path(parent, f"{child}_{i}").with_suffix(".dcm")
        shutil.copy(dicom_file, child)
        file_list.append(child)

    return file_list


def test_dicom_types(dicom_folder, capsys, tmp_path):
    sys.argv = [
        sys.argv[0],
        str(tmp_path),
    ]
    runpy.run_module("dicom_utils.cli.dicom_types", run_name="__main__", alter_sys=True)
    captured = capsys.readouterr()
    assert "1 - None - ORIGINAL|PRIMARY|AXIAL" in captured.out
    assert len(captured.out.split("\n")) == 2



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import re
from dataclasses import dataclass
from enum import Enum, auto
from typing import Any, Dict, Final, Iterable, Iterator, List, NamedTuple, Optional, TypeVar, Union, cast

from pydicom import DataElement
from pydicom.dataset import Dataset
from pydicom.multival import MultiValue
from pydicom.sequence import Sequence

from .tags import Tag


UNKNOWN: Final = -1


Dicom = Dataset
DicomAttributeSequence = Sequence


class DicomValueError(ValueError):
    pass


class DicomKeyError(ValueError):
    pass


def get_value(dcm: Dicom, tag: Tag, default: Any, try_file_meta: bool = False) -> Any:
    if tag in dcm:
        return dcm.get(tag).value
    elif try_file_meta and hasattr(dcm, "file_meta") and tag in dcm.file_meta:
        return dcm.file_meta.get(tag).value
    return default


def get_tag_values(tags: Iterable[Tag], dcm: Dicom) -> Dict[Tag, Any]:
    tags = {tag: get_value(dcm, tag, None) for tag in tags}
    return tags


# TODO: these would be better in dicom.py, but circular import issues
def iterate_view_codes(dcm: Dataset) -> Iterator[Dataset]:
    r"""Iterates over all view codes in an input."""
    view_code_sequence = (get_value(dcm, Tag.ViewCodeSequence, []) or []) if isinstance(dcm, Dataset) else dcm
    for view_code in view_code_sequence:
        yield view_code


def iterate_view_modifier_codes(dcm: Dataset) -> Iterator[Dataset]:
    r"""Iterates over all view modifier codes in an input."""
    # view modifier code can be at top level of dicom, or nested in view codes
    modifier_sequence = (get_value(dcm, Tag.ViewModifierCodeSequence, []) or []) if isinstance(dcm, Dataset) else dcm
    for modifier in modifier_sequence:
        yield modifier
    for view_code in iterate_view_codes(dcm):
        for modifier in iterate_view_modifier_codes(view_code):
            yield modifier


def iterate_shared_functional_groups(dcm: Dataset) -> Iterator[Dataset]:
    r"""Iterates over all view modifier codes in an input."""
    # view modifier code can be at top level of dicom, or nested in view codes
    functional_group_seq = (
        (get_value(dcm, Tag.SharedFunctionalGroupsSequence, []) or []) if isinstance(dcm, Dataset) else dcm
    )
    for seq in functional_group_seq:
        yield seq


T = TypeVar("T")


class EnumMixin(Enum):
    def __bool__(self) -> bool:
        return not self.is_unknown

    def __repr__(self) -> str:
        name = self.simple_name
        return f"{self.__class__.__name__}({name})"

    def __add__(self: T, other: T) -> T:
        return self or other

    def __mul__(self: T, other: T) -> T:
        return self or other

    @property
    def is_unknown(self) -> bool:
        return self.value == UNKNOWN

    @property
    def simple_name(self) -> str:
        return self.name.lower().replace("_", " ")

    @staticmethod
    def get_required_tags() -> List[Tag]:
        raise NotImplementedError("get_required_tags() has not been implemented for this class")


class MammogramType(EnumMixin):
    r"""Enum over the subcategories of mammograms.

    Supports the following ordering: ``TOMO < FFDM < SYNTH < SFM < UNKNOWN``
    """

    UNKNOWN = 0
    TOMO = 1
    FFDM = 2
    SYNTH = 3
    SFM = 4

    def __lt__(self, other: "MammogramType") -> bool:
        return self.is_preferred_to(other)

    def __le__(self, other: "MammogramType") -> bool:
        return self.is_preferred_to(other) or self == other

    def __bool__(self) -> bool:
        return self != MammogramType.UNKNOWN

    # TODO: MammogramType ordering uses the convention that x < y means x is more preferred than y
    # We may want to change this to x > y means x is more preferred than y. This would be more intuitive
    # but would mean `sorted(vals)` would return the values in the opposite order.
    def is_preferred_to(self, other: "MammogramType") -> bool:
        r"""Returns whether the current mammogram type is preferred to another.

        Args:
            other: The other mammogram type.

        Returns:
            Whether the current mammogram type is preferred to the other.
        """
        if self.is_unknown:
            return False
        elif other.is_unknown:
            return True
        return self.value < other.value

    @staticmethod
    def get_best(types: List["MammogramType"]) -> "MammogramType":
        r"""Returns the best mammogram type from a list of types."""
        if not types:
            raise ValueError("types must not be empty")
        return min(types)

    @property
    def is_unknown(self) -> bool:
        return not bool(self)

    @staticmethod
    def get_required_tags() -> List[Tag]:
        return [Tag.ImageType, Tag.Modality, Tag.NumberOfFrames, Tag.SeriesDescription]

    @staticmethod
    def from_dicom(dcm: Dicom, is_sfm: bool = False, ignore_modality: bool = False) -> "MammogramType":
        if (modality := get_value(dcm, Tag.Modality, None)) not in (None, "MG") and not ignore_modality:
            raise DicomValueError(f"Expected modality=MG, found {modality}")

        # if DICOM is a 3D volume, it must be tomo
        if dcm.get("NumberOfFrames", 1) > 1:
            return MammogramType.TOMO

        img_type = ImageType.from_dicom(dcm)
        pixels = img_type.pixels.lower()
        exam = img_type.exam.lower()
        flavor = (img_type.flavor or "").lower()
        extras = img_type.extras
        machine = get_value(dcm, Tag.ManufacturerModelName, "").lower()
        series_description = get_value(dcm, Tag.SeriesDescription, "").lower()

        # if fields 1 and 2 were missing, we know nothing
        if not img_type.pixels and img_type.exam:
            return MammogramType.UNKNOWN

        # very solid rules
        if is_sfm:
            return MammogramType.SFM
        if series_description and ("s-view" in series_description or "c-view" in series_description):
            return MammogramType.SYNTH
        if "original" in pixels:
            return MammogramType.FFDM

        # ok rules
        if extras is not None and any("generated_2d" in x.lower() for x in extras):
            return MammogramType.SYNTH

        # not good rules
        if pixels == "derived" and exam == "primary" and machine == "fdr-3000aws" and flavor != "post_contrast":
            return MammogramType.SYNTH

        return MammogramType.FFDM

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        if self is self.UNKNOWN:
            return "unknown"
        elif self is self.FFDM:
            return "ffdm"
        elif self is self.SFM:
            return "sfm"
        elif self is self.SYNTH:
            return "s-view"
        elif self is self.TOMO:
            return "tomo"
        else:
            raise RuntimeError("unknown ImageType value")

    @classmethod
    def from_str(cls, s: str) -> "MammogramType":
        if "tomo" in s:
            return cls.TOMO
        elif "view" in s or "synth" in s:
            return cls.SYNTH
        elif "2d" in s or "ffdm" in s:
            return cls.FFDM
        elif "sfm" in s:
            return cls.SFM
        else:
            return cls.UNKNOWN


@dataclass(frozen=True)
class ImageType:
    """Container for DICOM metadata fields related to the image type.

    Contains the following attributes:

        * ``"pixels"`` - First element of the ImageType field.
        * ``"exam"`` - Second element of the ImageType field.
        * ``"flavor"`` - Third element of the ImageType field.
        * ``"extras"`` - Additional ImageType fields if available
    """

    pixels: str
    exam: str
    flavor: Optional[str] = None
    extras: Optional[List[str]] = None

    def __bool__(self) -> bool:
        return bool(self.pixels) and bool(self.exam)

    def __contains__(self, val: Any) -> bool:
        return val == self.pixels or val == self.exam or val == self.flavor or val in (self.extras or [])

    def simple_repr(self) -> str:
        s = "|".join(x for x in (self.pixels, self.exam))
        if self.flavor is not None:
            s += f"|{self.flavor}" if self.flavor else "|''"
        for elem in self.extras or []:
            if not elem or elem.isdigit():
                continue
            s += f"|{elem}"
        return s

    @classmethod
    def from_dicom(cls, dcm: Dicom) -> "ImageType":
        result: Dict[str, Any] = {}

        if Tag.ImageType not in dcm.keys():
            return cls("", "", **result)

        # fields 1 and 2 should always be present
        image_type = cast(List[str], dcm[Tag.ImageType].value)
        pixels, exam = image_type[:2]
        result["pixels"] = pixels
        result["exam"] = exam

        # there might be a field 3
        if len(image_type) >= 3:
            flavor = image_type[2]
            result["flavor"] = flavor

        if len(image_type) >= 4:
            result["extras"] = image_type[3:]

        assert "pixels" in result.keys()
        assert "exam" in result.keys()
        return cls(**result)


class PhotometricInterpretation(Enum):
    r"""Enumeration of PhotometricInterpretation values under the DICOM
    standard. Values pulled from:

        https://dicom.innolitics.com/ciods/rt-dose/image-pixel/00280004
    """

    UNKNOWN = 0
    MONOCHROME1 = 1
    MONOCHROME2 = 2
    PALETTE_COLOR = auto()
    RGB = auto()
    HSV = auto()
    ARGB = auto()
    CMYK = auto()
    YBR_FULL = auto()
    YBR_FULL_422 = auto()
    YBR_PARTIAL_422 = auto()
    YBR_PARTIAL_420 = auto()
    YBR_ICT = auto()
    YBR_RCT = auto()

    def __bool__(self) -> bool:
        return self != PhotometricInterpretation.UNKNOWN

    @property
    def is_monochrome(self) -> bool:
        return 1 <= self.value <= 2

    @property
    def num_channels(self) -> int:
        return 1 if self.is_monochrome else 3

    @property
    def is_inverted(self) -> bool:
        return self == PhotometricInterpretation.MONOCHROME1

    @classmethod
    def from_str(cls, string: str) -> "PhotometricInterpretation":
        string = string.upper()
        return getattr(cls, string, PhotometricInterpretation.UNKNOWN)

    @classmethod
    def from_dicom(cls, dcm: Dicom) -> "PhotometricInterpretation":
        val = dcm.get(Tag.PhotometricInterpretation, None)
        return PhotometricInterpretation.UNKNOWN if val is None else cls.from_str(val.value)


class Laterality(EnumMixin):
    UNKNOWN = UNKNOWN

    NONE = 0
    LEFT = 1
    RIGHT = 2
    BILATERAL = 3

    @property
    def opposite(self) -> "Laterality":
        if self == Laterality.LEFT:
            return Laterality.RIGHT
        elif self == Laterality.RIGHT:
            return Laterality.LEFT
        return Laterality.UNKNOWN

    @property
    def is_unilateral(self) -> bool:
        return self in (Laterality.LEFT, Laterality.RIGHT)

    @property
    def is_unknown_or_none(self) -> bool:
        return self in (Laterality.UNKNOWN, Laterality.NONE)

    @staticmethod
    def get_required_tags() -> List[Tag]:
        return [
            Tag.Laterality,
            Tag.ImageLaterality,
            Tag.FrameLaterality,
            Tag.SharedFunctionalGroupsSequence,
            Tag.PatientOrientation,
        ]

    @classmethod
    def from_str(cls, string: str) -> "Laterality":
        string = string.strip().lower()
        if string == "none":
            return cls.NONE
        if "bi" in string:  # TODO what other patterns describe bilateral?
            return cls.BILATERAL
        if "r" in string or "d" in string:
            return cls.RIGHT
        if "l" in string or "e" in string:
            return cls.LEFT
        return cls.UNKNOWN

    @classmethod
    def from_tags(cls, tags: Dict[int, Any]) -> "Laterality":
        # Take subset of 'tags' so that unit tests will fail if we don't maintain get_required_tags()
        tags = {k: v for k, v in tags.items() if k in cls.get_required_tags()}

        # first try reading ImageLaterality
        laterality = tags.get(Tag.ImageLaterality, "")

        # next try reading Laterality
        laterality = laterality or tags.get(Tag.Laterality, "")

        # fall back to Tag.FrameLaterality
        if not laterality:
            try:
                laterality = (
                    tags.get(Tag.SharedFunctionalGroupsSequence)[0]  # type: ignore
                    .get(Tag.FrameAnatomySequence)
                    .value[0]
                    .get(Tag.FrameLaterality)
                    .value
                )
            except Exception:
                pass

        # TODO is there a DICOM value for bilateral?
        laterality = laterality.strip().lower() if laterality else ""
        if laterality == "l":
            return cls.LEFT
        elif laterality == "r":
            return cls.RIGHT
        else:
            return cls.UNKNOWN

    @classmethod
    def from_patient_orientation(cls, value: Union[str, List[str]]) -> "Laterality":
        if isinstance(value, str):
            return cls.from_str(value)

        for item in value:
            parts = set(str(item))
            if "L" in parts:
                return cls.RIGHT
            elif "R" in parts:
                return cls.LEFT
            elif result := cls.from_str(item):
                return result
        return cls.UNKNOWN

    @classmethod
    def from_dicom(cls, dcm: Dicom) -> "Laterality":
        return cls.from_tags({int(tag): value for tag, value in get_tag_values(cls.get_required_tags(), dcm).items()})

    @classmethod
    def from_case_notes(cls, notes: str) -> "Laterality":
        notes = notes.lower()
        if "left" in notes:
            return cls.LEFT
        elif "right" in notes:
            return cls.RIGHT
        elif "bilateral" in notes:
            return cls.BILATERAL
        else:
            return cls.UNKNOWN

    @property
    def short_str(self) -> str:
        if self == Laterality.LEFT:
            return "l"
        elif self == Laterality.RIGHT:
            return "r"
        return ""

    def __add__(self, other: "Laterality") -> "Laterality":
        r"""Reduces two :class:`Laterality` inputs according to the following rules:
        * ANY + BILATERAL -> BILATERAL
        * LEFT + RIGHT -> BILATERAL
        * LEFT + (UNKNOWN/NONE) -> LEFT
        * RIGHT + (UNKNOWN/NONE) -> RIGHT
        * NONE + NONE -> NONE
        * UNKOWN + UNKOWN -> UNKNOWN
        """
        # either input is unknown, fall back to parent __add__
        if self.is_unknown or other.is_unknown:
            return cast(Laterality, super().__add__(other))

        # if either input is bilateral, output should be bilateral
        elif self == Laterality.BILATERAL or other == Laterality.BILATERAL:
            return Laterality.BILATERAL

        # if both inputs are unilateral complements, reduce to bilateral
        # if either inputs is a unilateral non-complements, return the unilateral laterality
        elif self.is_unilateral or other.is_unilateral:
            return (
                Laterality.BILATERAL
                if self.is_unilateral and other.is_unilateral and self != other
                else self if self.is_unilateral else other
            )

        # only remaining possibility
        return Laterality.NONE

    def reduce(self, *other: "Laterality") -> "Laterality":
        return sum((self, *other), Laterality.UNKNOWN)


CC_STRINGS: Final = {"cranio-caudal", "caudal-cranial"}
ML_STRINGS: Final = {"medio-lateral", "medial-lateral"}
LM_STRINGS: Final = {"latero-medial", "lateral-medial"}
MLO_STRINGS: Final = {pattern for s in ML_STRINGS for pattern in (f"{s} oblique", f"oblique {s}")}
LMO_STRINGS: Final = {pattern for s in LM_STRINGS for pattern in (f"{s} oblique", f"oblique {s}")}
XCCL_STRINGS: Final = {s + " exaggerated laterally" for s in CC_STRINGS}
XCCM_STRINGS: Final = {s + " exaggerated medially" for s in CC_STRINGS}
AT_STRINGS: Final = {"axillary tail"}
CV_STRINGS: Final = {"cleavage view", "valley-view"}


class ViewPosition(EnumMixin):
    UNKNOWN = UNKNOWN

    XCCL = auto()
    XCCM = auto()
    CC = auto()
    MLO = auto()
    ML = auto()
    LMO = auto()
    LM = auto()
    AT = auto()
    CV = auto()

    @staticmethod
    def get_required_tags() -> List[Tag]:
        return [Tag.ViewPosition, Tag.ViewCodeSequence]

    @classmethod
    def from_str(cls, string: str, strict: bool = False) -> "ViewPosition":
        string = string.strip().lower()

        # first try strict patterns
        strict_mapping = {
            cls.CC: CC_STRINGS,
            cls.MLO: MLO_STRINGS,
            cls.ML: ML_STRINGS,
            cls.LM: LM_STRINGS,
            cls.LMO: LMO_STRINGS,
            cls.XCCL: XCCL_STRINGS,
            cls.XCCM: XCCM_STRINGS,
            cls.AT: AT_STRINGS,
        }
        for value, keywords in strict_mapping.items():
            if string in keywords:
                return value
            elif value.simple_name == string:
                return value

        if strict:
            return cls.UNKNOWN

        # loose patterns
        for choice in cls:
            keyword = choice.simple_name
            if keyword and keyword in string:
                return choice
        return cls.UNKNOWN

    @classmethod
    def from_tags(cls, tags: Dict[int, Any]) -> "ViewPosition":
        # Take subset of 'tags' so that unit tests will fail if we don't maintain get_required_tags()
        tags = {k: v for k, v in tags.items() if k in cls.get_required_tags()}
        view_position = cls.from_view_position_tag(tags.get(Tag.ViewPosition, None))
        return (
            view_position
            if view_position is not cls.UNKNOWN
            else cls.from_view_code_sequence_tag(tags.get(Tag.ViewCodeSequence, None))
        )

    @classmethod
    def from_view_position_tag(cls, view_position: Optional[str]) -> "ViewPosition":
        if isinstance(view_position, str):
            return cls.from_str(view_position)
        return cls.UNKNOWN

    @classmethod
    def from_view_code_sequence_tag(cls, view_code_sequence: Optional[DataElement]) -> "ViewPosition":
        for view_code in view_code_sequence or []:
            meaning = view_code.get("CodeMeaning", None)
            if isinstance(meaning, str):
                return cls.from_str(meaning, strict=True)
        return cls.UNKNOWN

    @classmethod
    def from_patient_orientation(cls, value: Union[str, List[str]]) -> "ViewPosition":
        if isinstance(value, str):
            return cls.from_str(value)

        for item in value:
            if item in ("FL", "FR"):
                return cls.MLO
            elif item in ("R", "L"):
                return cls.CC
            elif result := cls.from_str(item):
                return result
        return cls.UNKNOWN

    @classmethod
    def from_dicom(cls, dcm: Dicom) -> "ViewPosition":
        def from_code(code: Dataset) -> ViewPosition:
            return cls.from_str(get_value(code, Tag.CodeMeaning, ""), strict=True)

        view_position = cls.from_str(get_value(dcm, Tag.ViewPosition, ""))
        view_codes = {result for code in iterate_view_codes(dcm) if (result := from_code(code))}
        view_modifier_codes = {result for code in iterate_view_modifier_codes(dcm) if (result := from_code(code))}
        candidates = view_codes.union(view_modifier_codes).union({view_position})
        return sorted(candidates, key=lambda x: x.value)[-1]

    @property
    def short_str(self) -> str:
        if self == self.__class__.UNKNOWN:
            return ""
        return self.name.lower()

    @property
    def is_standard_view(self) -> bool:
        return self in {self.CC, self.MLO}

    @property
    def is_mlo_like(self) -> bool:
        return self in {self.MLO, self.ML, self.LMO, self.LM}

    @property
    def is_cc_like(self) -> bool:
        return self in {self.CC, self.XCCL, self.XCCM}


class MammogramView(NamedTuple):
    laterality: Laterality = Laterality.UNKNOWN
    view: ViewPosition = ViewPosition.UNKNOWN

    @classmethod
    def create(cls, laterality: Optional[Laterality] = None, view: Optional[ViewPosition] = None) -> "MammogramView":
        return cls(
            laterality or Laterality.UNKNOWN,
            view or ViewPosition.UNKNOWN,
        )

    @classmethod
    def from_dicom(cls, dcm: Dicom) -> "MammogramView":
        return cls.create(
            Laterality.from_dicom(dcm),
            ViewPosition.from_dicom(dcm),
        )

    @property
    def is_standard_mammo_view(self) -> bool:
        r"""Checks if this record corresponds to a standard mammography view.
        Standard mammography views are the MLO and CC views.
        """
        return self.view.is_standard_view

    @staticmethod
    def get_required_tags() -> List[Tag]:
        return [*Laterality.get_required_tags(), *ViewPosition.get_required_tags()]

    @property
    def is_mlo_like(self) -> bool:
        return self.view.is_mlo_like

    @property
    def is_cc_like(self) -> bool:
        return self.view.is_cc_like


# Matches floats, incluidng exponential notation
FLOAT_PATTERN = r"\d+\.?\d*(?:[e\-\d]+)?"

PIXEL_SPACING_RE = re.compile(rf"({FLOAT_PATTERN})[^\d.]+({FLOAT_PATTERN})")


@dataclass(frozen=True)
class PixelSpacing:
    r"""Represents detector pixel spacing in mm."""

    row: float
    col: float

    @classmethod
    def from_str(cls, string: str) -> "PixelSpacing":
        # value will be of from [row, col] in mm
        match = PIXEL_SPACING_RE.search(string)
        if match:
            try:
                values = tuple(float(x) for x in match.groups())
                row, col = values
                return cls(row, col)
            except ValueError:
                pass
        raise ValueError(f"Failed to parse PixelSpacing from {string}")

    @classmethod
    def from_tags(cls, tags: Dict[Tag, Any]) -> "PixelSpacing":
        for tag, spacing in tags.items():
            try:
                if isinstance(spacing, str):
                    return cls.from_str(spacing)
                elif isinstance(spacing, MultiValue):
                    row, col = tuple(float(x) for x in spacing)
                    return cls(row, col)
            except ValueError:
                pass
        raise RuntimeError("Failed to create PixelSpacing from DICOM")

    @classmethod
    def from_dicom(cls, dcm: Dicom) -> "PixelSpacing":
        return cls.from_tags({tag: value for tag, value in get_tag_values(cls.get_required_tags(), dcm).items()})

    @staticmethod
    def get_required_tags() -> List[Tag]:
        return [Tag.ImagerPixelSpacing, Tag.PixelSpacing]


__all__ = [
    "Dicom",
    "ImageType",
    "PhotometricInterpretation",
    "EnumMixin",
    "Laterality",
    "ViewPosition",
    "MammogramView",
    "PixelSpacing",
]



import numpy as np
import pytest

from dicom_utils.visualize import *


@pytest.mark.parametrize(
    "images, expected",
    [
        ([np.array([[[[0]], [[0]], [[0]]]])], np.array([[[[0]], [[0]], [[0]]]])),
        ([np.array([[[[0]], [[1]], [[2]]]])], np.array([[[[0]], [[1]], [[2]]]])),
        (
            [np.array([[[[0]], [[0]], [[0]]]]), np.array([[[[1]], [[1]], [[1]]]])],
            np.array([[[[0, 1]], [[0, 1]], [[0, 1]]]]),
        ),
        (
            [np.array([[[[0]], [[0]], [[0]]]]), np.array([[[[1]], [[1]], [[1]]]]), np.array([[[[2]], [[2]], [[2]]]])],
            np.array([[[[0, 1], [2, 0]], [[0, 1], [2, 0]], [[0, 1], [2, 0]]]]),
        ),
        (
            [
                np.array([[[[7, 0, 0]], [[0, 5, 0]], [[0, 0, 8]]]]),
                np.array([[[[1]], [[1]], [[1]]]]),
                np.array([[[[2]], [[2]], [[2]]]]),
            ],
            np.array(
                [
                    [
                        [[7, 0, 0, 1, 0, 0], [2, 0, 0, 0, 0, 0]],
                        [[0, 5, 0, 1, 0, 0], [2, 0, 0, 0, 0, 0]],
                        [[0, 0, 8, 1, 0, 0], [2, 0, 0, 0, 0, 0]],
                    ]
                ]
            ),
        ),
    ],
)
def test_to_collage(images, expected) -> None:
    collage = to_collage(images)
    assert (expected == collage).all()
    assert expected.shape == collage.shape


def test_to_collage_num_image_assertion() -> None:
    with pytest.raises(AssertionError, match="There must be at least one image."):
        to_collage([])


@pytest.mark.parametrize(
    "ndarrays",
    [
        ([np.array([])]),
        ([np.array([[[0]]]), np.array([])]),
    ],
)
def test_to_collage_image_shape_assertion(ndarrays) -> None:
    with pytest.raises(AssertionError, match="The images must have 4 dimensions."):
        to_collage(ndarrays)


@pytest.mark.parametrize(
    "data, bbox",
    [
        ([1, 1, 2, 2], [0, 0, 2, 2]),
        ([6, 6, 6, 10], [2, 2, 10, 10]),
    ],
)
def test_dicom_circle_to_bbox(data, bbox) -> None:
    predicted_bbox = dicom_circle_to_bbox(data)
    assert list(predicted_bbox) == bbox


@pytest.mark.parametrize(
    "data, bbox",
    [
        ([0, 1, 2, 1, 1, 0, 1, 2], [0, 0, 2, 2]),
    ],
)
def test_dicom_ellipse_to_bbox(data, bbox) -> None:
    predicted_bbox = dicom_ellipse_to_bbox(data)
    assert list(predicted_bbox) == bbox


@pytest.mark.parametrize(
    "form, data",
    [
        (Form("ELLIPSE"), [0] * 8),
        (Form("CIRCLE"), [0] * 4),
    ],
)
def test_dicom_trace_to_bbox(form, data):
    dicom_trace_to_bbox(data, form)


@pytest.mark.parametrize(
    "data, expected_shape",
    [(np.array([0] * 60).reshape((3, 4, 5)), (4, 5, 3))],
)
def test_channel_swaps(data, expected_shape):
    intermediate = chw_to_hwc(data)
    assert intermediate.shape == expected_shape
    assert hwc_to_chw(intermediate).shape == data.shape


@pytest.mark.parametrize(
    "data, expected_shape",
    [(np.array([1] * 20).reshape((1, 4, 5)), (1, 3, 4, 5))],
)
def test_to_rgb(data, expected_shape):
    assert to_rgb(data).shape == expected_shape


@pytest.mark.parametrize(
    "data, expected",
    [
        ([0, 1], [0, 255]),
        ([1, 2], [0, 255]),
        ([0, 1, 2], [0, 128, 255]),
        ([0, 0.5, 1], [0, 128, 255]),
    ],
)
def test_to_8bit(data, expected):
    output = to_8bit(np.array(data))
    assert (output == np.array(expected)).all()


@pytest.mark.parametrize(
    "data_a, data_b, expected",
    [([0, 0], [0, 1], 1)],
)
def test_distance(data_a, data_b, expected):
    assert distance(data_a, data_b) == expected


@pytest.mark.parametrize(
    "data, expected",
    [
        ([[0, 1]], [[0, 1]]),
        ([[0, 1], [1, 2]], [[0, 1], [1, 2]]),
        ([[0, 1, 8, 9], [8, 9, 3, 4]], [[0, 1, 8, 9, 8, 9, 3, 4]]),
        ([[0, 1, 2, 3], [8, 9, 3, 4]], [[0, 1, 2, 3], [8, 9, 3, 4]]),
    ],
)
def test_group_polylines(data, expected):
    data = [GraphicItem(x, Form.POLYLINE) for x in data]
    expected = [GraphicItem(x, Form.POLYLINE) for x in expected]
    assert list(group_polylines(data)) == expected



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from os import PathLike
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Hashable,
    Iterable,
    Iterator,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from registry import Registry, bind_relevant_kwargs

from .collection import RecordCollection
from .group import Grouper
from .record import RECORD_REGISTRY, SupportsPatientID, SupportsStudyDate, SupportsStudyUID


logger = logging.getLogger(__name__)
NAME_REGISTRY = Registry("names", bound=Type[Callable[[Hashable, RecordCollection, int, int], str]])
K = TypeVar("K", bound=Hashable)


class CaseRenamer(ABC, Generic[K]):
    @abstractmethod
    def __call__(self, key: K, collection: RecordCollection, index: int, total: int) -> str:
        raise NotImplementedError

    @classmethod
    def num_leading_zeros(cls, total: int) -> int:
        return len(str(total))

    @classmethod
    def add_leading_zeros(cls, index: int, total: int) -> str:
        return str(index).zfill(cls.num_leading_zeros(total))


@NAME_REGISTRY(name="key")
@dataclass
class UseKeyRenamer(CaseRenamer):
    prefix: str = ""

    def __call__(self, key: Hashable, collection: RecordCollection, index: int, total: int) -> str:
        key = str(key) if key is not None else "???-{index}"
        return f"{self.prefix}{key}"


@NAME_REGISTRY(name="consecutive")
@dataclass
class ConsecutiveNamer(CaseRenamer[K]):
    prefix: str = "Case-"
    start: int = 1

    def __call__(self, key: K, collection: RecordCollection, index: int, total: int) -> str:
        return f"{self.prefix}{self.add_leading_zeros(index, total)}"


@NAME_REGISTRY(name="patient-id")
@dataclass
class PatientIDNamer(CaseRenamer[Optional[str]]):
    prefix: str = "Patient-"

    def __call__(self, key: Optional[str], collection: RecordCollection, index: int, total: int) -> str:
        pid = self.get_patient_id(collection)
        return f"{self.prefix}{pid}" if pid else ""

    @classmethod
    def get_patient_id(cls, collection: RecordCollection) -> Optional[str]:
        pids = {pid for rec in collection if isinstance(rec, SupportsPatientID) and (pid := rec.PatientID) is not None}
        pid = next(iter(pids)) if len(pids) == 1 else None
        # Replace slashes with underscores because they are not allowed in filenames
        if pid is not None:
            pid = pid.replace("/", "_")
        return pid


@NAME_REGISTRY(name="study-date")
@dataclass
class StudyDateNamer(CaseRenamer):
    prefix: str = "Date-"
    year_only: bool = False

    def __call__(self, key: Optional[str], collection: RecordCollection, index: int, total: int) -> str:
        date = self.get_study_date(collection, self.year_only)
        return f"{self.prefix}{date}" if date is not None else ""

    @classmethod
    def get_study_date(cls, collection: RecordCollection, year_only: bool = False) -> Optional[str]:
        dates = {
            str(date)
            for rec in collection
            if isinstance(rec, SupportsStudyDate)
            and (date := (rec.StudyYear if year_only else rec.StudyDate)) is not None
        }
        return next(iter(dates)) if len(dates) == 1 else None


@NAME_REGISTRY(name="study-uid")
@dataclass
class StudyIDNamer(CaseRenamer[Optional[str]]):
    prefix: str = "Study-"
    truncate: Optional[int] = 4
    strip_period: bool = True
    patient_namer: Optional[str] = None
    date_namer: Optional[str] = None

    _patient_namer: Optional[PatientIDNamer] = field(default=None, init=False)
    _date_namer: Optional[StudyDateNamer] = field(default=None, init=False)

    def __post_init__(self):
        if self.patient_namer:
            self._patient_namer = cast(
                PatientIDNamer,
                NAME_REGISTRY.get(self.patient_namer).instantiate_with_metadata().fn,
            )
            assert isinstance(self._patient_namer, PatientIDNamer), type(self._patient_namer)
        if self.date_namer:
            self._date_namer = cast(
                StudyDateNamer,
                NAME_REGISTRY.get(self.date_namer).instantiate_with_metadata().fn,
            )
            assert isinstance(self._date_namer, StudyDateNamer), type(self._date_namer)

    def __call__(self, key: Optional[str], collection: RecordCollection, index: int, total: int) -> str:
        uid = self.get_study_uid(collection)
        if not uid:
            return ""

        uid = self._strip(uid)
        uid = self._truncate(uid)
        pid = self._call_patient_namer(key, collection, index, total) if self._patient_namer is not None else ""
        date = self._call_date_namer(key, collection, index, total) if self._date_namer is not None else ""

        parts = [s for s in (pid, date, f"{self.prefix}{uid}") if s]
        return "_".join(parts)

    @classmethod
    def get_study_uid(cls, collection: RecordCollection) -> Optional[str]:
        uids = {
            uid for rec in collection if isinstance(rec, SupportsStudyUID) and (uid := rec.StudyInstanceUID) is not None
        }
        return next(iter(uids)) if len(uids) == 1 else None

    def _truncate(self, uid: str) -> str:
        if self.truncate is not None:
            assert self.truncate > 0
            uid = uid[-self.truncate :]
        return uid

    def _strip(self, uid: str) -> str:
        if self.strip_period:
            uid = uid.replace(".", "")
        return uid

    def _call_patient_namer(self, key: Optional[str], collection: RecordCollection, index: int, total: int) -> str:
        assert self._patient_namer is not None
        return self._patient_namer(key, collection, index, total)

    def _call_date_namer(self, key: Optional[str], collection: RecordCollection, index: int, total: int) -> str:
        assert self._date_namer is not None
        return self._date_namer(key, collection, index, total)


class Input:
    r"""Input pipeline for discovering files, creating :class:`FileRecord`s, grouping records, and naming
    each group. :class:`Input` is an iterable over the deterministically sorted record groups and their
    corresponding group name.

    Args:
        sources:
            A directory or iterable of directories from which to read

        records:
            An iterable of names for :class:`FileRecord` subclasses registered in ``RECORD_REGISTRY``.
            Defaults to all registered :class:`FileRecord` subclasses.

        groups:
            An iterable of names for grouping functions registered in ``GROUP_REGISTRY``.
            Defaults to grouping by StudyInstanceUID.

        helpers:
            An iterable of names for :class:`RecordHelper` subclasses registered in ``HELPER_REGISTRY``.
            By default no helpers will be used.

        filters:
            An iterable of names for :class:`RecordFilter` subclasses registered in ``FILTER_REGISTRY``.
            By default no filters will be used.

    """

    def __init__(
        self,
        sources: Union[PathLike, Iterable[PathLike]],
        records: Optional[Iterable[str]] = None,
        groups: Iterable[str] = ["patient-id", "study-uid"],
        helpers: Iterable[str] = [],
        namers: Iterable[str] = ["patient-id", "study-uid"],
        filters: Iterable[str] = [],
        **kwargs,
    ):
        if records is None:
            self.records = RECORD_REGISTRY.available_keys()
        else:
            self.records = records
        namers = list(namers)
        groups = list(groups)
        helpers = list(helpers)

        self.grouper = bind_relevant_kwargs(Grouper, groups=groups, helpers=helpers, **kwargs)()
        self.namers = [NAME_REGISTRY.get(n).instantiate_with_metadata() for n in namers]
        if len(namers) != len(groups):
            raise ValueError("Number of namers {namers} should match number of groups {groups}")

        # scan sources and build a RecordCollection with every valid file found
        sources = (Path(p) for p in ([sources] if isinstance(sources, PathLike) else sources))
        collection = RecordCollection.create(
            sources, record_types=self.records, helpers=helpers, filters=filters, **kwargs
        )

        # apply groupers to generate a dict of key -> group pairs
        grouped_collections = self.grouper(collection)

        # apply namers to generate a dict of (group name) -> group pairs
        self.cases: Dict[Tuple[str, ...], RecordCollection] = {}
        for i, group_key in enumerate(sorted(grouped_collections.keys(), key=cast(Any, self.sort_key))):
            group = grouped_collections[group_key]
            group_key = (group_key,) if not isinstance(group_key, tuple) else group_key
            key = tuple(namer(k, group, i + 1, len(grouped_collections)) for namer, k in zip(self.namers, group_key))
            self.cases[key] = group

    def __len__(self) -> int:
        return len(self.cases)

    def __iter__(self) -> Iterator[Tuple[Tuple[str, ...], RecordCollection]]:
        r"""Iterates over pairs of named groups and the :class:`RecordCollection` containing that group."""
        for k, v in self.cases.items():
            yield k, v

    @staticmethod
    def sort_key(key: Optional[Tuple[Hashable, ...]]) -> Tuple[str, ...]:
        if key is None:
            return ("None",)
        return tuple(str(x) for x in key)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from dataclasses import dataclass
from typing import Optional

import pytest

from dicom_utils.container.protocols import (
    SupportsManufacturer,
    SupportsPatientAge,
    SupportsPatientID,
    SupportsStudyUID,
    SupportsUID,
)


class TestSupportsStudyUID:
    @pytest.fixture
    def factory(self):
        @dataclass
        class Impl(SupportsStudyUID):
            StudyInstanceUID: Optional[str] = None

        return Impl

    @pytest.mark.parametrize(
        "uid1,uid2,exp",
        [
            pytest.param("1.2.3", "1.2.3", True),
            pytest.param(None, "1.2.3", False),
            pytest.param("1.2.3", None, False),
        ],
    )
    def test_same_study_as(self, factory, uid1, uid2, exp):
        s1 = factory(uid1)
        s2 = factory(uid2)
        assert s1.same_study_as(s2) == exp


class TestSupportsPatientID:
    @pytest.fixture
    def factory(self):
        @dataclass
        class Impl(SupportsPatientID):
            PatientID: Optional[str] = None
            PatientName: Optional[str] = None

        return Impl

    @pytest.mark.parametrize(
        "id1,name1,id2,name2,exp",
        [
            pytest.param("P1", None, "P1", None, True),
            pytest.param("P1", None, "P2", None, False),
            pytest.param("P1", None, None, None, False),
            pytest.param(None, None, None, None, False),
            pytest.param("P1", "John Doe", "P2", "John Doe", False),
            pytest.param("P1", "John A. Doe", "P1", "John Alex Doe", True),
        ],
    )
    def test_same_patient_as(self, factory, id1, name1, id2, name2, exp):
        s1 = factory(id1, name1)
        s2 = factory(id2, name2)
        assert s1.same_patient_as(s2) == exp


class TestSupportsUID:
    @pytest.fixture
    def factory(self):
        @dataclass
        class Impl(SupportsUID):
            SOPInstanceUID: Optional[str] = None
            SeriesInstanceUID: Optional[str] = None

        return Impl

    @pytest.mark.parametrize(
        "sop,series,prefer_sop,exp",
        [
            pytest.param("1.2.3", None, True, "1.2.3"),
            pytest.param(None, "1.2.3", True, "1.2.3"),
            pytest.param("1.2.3", "2.3.4", True, "1.2.3"),
            pytest.param("1.2.3", "2.3.4", False, "2.3.4"),
            pytest.param(None, None, True, None, marks=pytest.mark.xfail(raises=AttributeError)),
        ],
    )
    def test_get_uid(self, factory, sop, series, prefer_sop, exp):
        s = factory(sop, series)
        assert s.get_uid(prefer_sop) == exp

    @pytest.mark.parametrize(
        "sop1,series1,sop2,series2,exp",
        [
            pytest.param("1.2.3", None, "1.2.3", None, True),
            pytest.param("1.2.3", None, "2.3.4", None, False),
            pytest.param("1.2.3", "1.2.3", "2.3.4", "1.2.3", False),
            pytest.param(None, "1.2.3", None, "1.2.3", True),
            pytest.param(None, None, None, None, False),
        ],
    )
    def test_same_patient_as(self, factory, sop1, series1, sop2, series2, exp):
        s1 = factory(sop1, series1)
        s2 = factory(sop2, series2)
        assert s1.same_uid_as(s2) == exp


class TestSupportsManufacturer:
    @pytest.fixture
    def factory(self):
        @dataclass
        class Impl(SupportsManufacturer):
            Manufacturer: Optional[str] = None
            ManufacturerModelName: Optional[str] = None
            ManufacturerModelNumber: Optional[str] = None

        return Impl

    @pytest.mark.parametrize(
        "man,model,number,pattern,exp",
        [
            pytest.param(None, None, None, "foo", None),
            pytest.param("man", "model", "number", "m.*", "man"),
            pytest.param(None, "model", "number", "m.*", "model"),
            pytest.param("foo", "bar", "model_num", "m.*", "model_num"),
        ],
    )
    def test_search_manufacturer_info(self, factory, man, model, number, pattern, exp):
        s = factory(man, model, number)
        m = s.search_manufacturer_info(pattern)
        if exp is None:
            assert m is None
        else:
            assert m.group() == exp


class TestSupportsPatientAge:
    @pytest.fixture
    def factory(self):
        @dataclass
        class Impl(SupportsPatientAge):
            PatientAge: Optional[str] = None
            PatientBirthDate: Optional[str] = None

        return Impl

    @pytest.mark.parametrize(
        "age,exp",
        [
            pytest.param("095Y", 95),
            pytest.param("020Y", 20),
            pytest.param("99Y", 99),
            pytest.param("32", 32),
            pytest.param("", None),
            pytest.param(None, None),
        ],
    )
    def test_patient_age(self, factory, age, exp):
        s = factory(age)
        assert s.patient_age == exp



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from abc import ABC, abstractmethod
from copy import copy, deepcopy
from functools import partial
from inspect import signature
from itertools import product
from os import PathLike
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Optional, Type, TypeVar, Union, cast

import numpy as np
import pydicom
from numpy.random import default_rng
from pydicom import DataElement, Dataset, FileDataset, Sequence
from pydicom.data import get_testdata_file
from pydicom.datadict import dictionary_VR, keyword_dict
from pydicom.valuerep import VR
from registry import Registry

from .dicom import set_pixels
from .tags import Tag


T = TypeVar("T", bound="DicomFactory")
C = TypeVar("C", bound="ConcatFactory")
U = TypeVar("U")
Proto = Union[PathLike, FileDataset, str]
DEFAULT_PROTO = "CT_small.dcm"
FACTORY_REGISTRY = Registry("factories")


def args_from_dicom(func: Callable[..., U], dicom: FileDataset) -> Callable[..., U]:
    sig = signature(func)
    kwargs: Dict[str, Any] = {}
    for name in sig.parameters.keys():
        if not hasattr(Tag, name) or not hasattr(dicom, name):
            continue
        tag = getattr(Tag, name)
        kwargs[name] = dicom[tag].value
    return partial(func, **kwargs)


class BaseFactory(ABC):
    @abstractmethod
    def __call__(self, seed: Optional[int] = None, **kwargs) -> FileDataset:
        raise NotImplementedError

    @classmethod
    def data_element(cls, tag: Tag, value: Any, vr: Optional[Union[VR, str]] = None, **kwargs) -> DataElement:
        if vr is None:
            vr = cls.suggest_vr(tag, value)
        else:
            vr = VR(vr)
        return DataElement(tag, vr, value, **kwargs)

    @classmethod
    def pixel_array(
        cls,
        Rows: int,
        Columns: int,
        NumberOfFrames: int = 1,
        BitsStored: int = 16,
        BitsAllocated: int = 14,
        PhotometricInterpretation: str = "MONOCHROME2",
        seed: int = 42,
    ) -> np.ndarray:
        low = 0
        high = 2**BitsAllocated - 1
        channels = 1 if PhotometricInterpretation.startswith("MONOCHROME") else 3
        size = tuple(x for x in (channels, NumberOfFrames, Rows, Columns) if x > 1)
        rng = default_rng(seed)
        dtype = np.uint16 if BitsStored > 8 else np.uint8
        return rng.integers(low, high, size, dtype=dtype)

    @classmethod
    def random_uid(cls, length: int = 6, seed: int = 42) -> str:
        rng = default_rng(seed)
        return "".join(str(x) for x in rng.integers(0, 10, length))

    @classmethod
    def pixel_array_from_dicom(
        cls,
        dcm: FileDataset,
        seed: int = 42,
    ) -> np.ndarray:
        func = args_from_dicom(cls.pixel_array, dcm)
        return func(seed=seed)

    @classmethod
    def code_sequence(cls, *meanings: str) -> Sequence:
        codes: List[Dataset] = []
        for meaning in meanings:
            vc = Dataset()
            vc[Tag.CodeMeaning] = cls.data_element(Tag.CodeMeaning, meaning, "ST")
            codes.append(vc)
        return Sequence(codes)

    @classmethod
    def suggest_vr(cls, tag: Tag, value: Any) -> VR:
        name = tag.name
        if name in keyword_dict:
            return VR(dictionary_VR(keyword_dict[name]))
        return VR("ST")

    @classmethod
    def save_dicoms(cls, path: PathLike, dicoms: Iterable[FileDataset]) -> List[Path]:
        root = Path(path)
        results: List[Path] = []
        for i, dcm in enumerate(dicoms):
            path = Path(root, f"D{i}.dcm")
            dcm.save_as(path)
            results.append(path)
        return results


# NOTE: This class should not use any dicom-utils methods in its implementation, as it will
# be used in the testing of dicom-utils
@FACTORY_REGISTRY(name="dicom")
class DicomFactory(BaseFactory):
    r"""Factory class for creating DICOM objects for unit tests.

    Args:
        proto:
            A prototype DICOM on which defaults will be based. Can be a DICOM FileDataset object,
            a path to a DICOM file, or a string with a pydicom testdata file.

        seed:
            Seed for random number generation

        pixels:
            Whether to include pixel data in the generated DICOM. If False, ``dcm.PixelData`` will be set to ``None``.

    Keyword Args:
        Tag value overrides
    """

    dicom: FileDataset

    def __init__(
        self,
        proto: Proto = DEFAULT_PROTO,
        seed: int = 42,
        allow_nonproto_tags: bool = True,
        pixels: bool = True,
        **kwargs,
    ):
        self.seed = int(seed)
        self.rng = default_rng(self.seed)
        self.allow_nonproto_tags = allow_nonproto_tags
        self.pixels = pixels
        self.overrides = kwargs
        if isinstance(proto, (PathLike, str)):
            self.path = Path(proto)
            if not self.path.is_file():
                self.path = Path(cast(str, get_testdata_file(str(proto))))
            self.dicom = pydicom.dcmread(self.path)

        elif isinstance(proto, FileDataset):
            self.path = None
            self.dicom = proto

        else:
            raise TypeError(f"`proto` should be PathLike or FileDataset, found {type(proto)}")

    def __add__(self: T, other: T) -> T: ...

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.path})"

    def __enter__(self: T) -> T:
        result = copy(self)
        result.seed = self.seed
        result.rng = default_rng(self.seed)
        return result

    def __call__(self, seed: Optional[int] = None, pixels: Optional[bool] = None, **kwargs) -> FileDataset:
        self.rng if seed is None else default_rng(seed)
        dcm = deepcopy(self.dicom)

        # set overrides
        overrides = {**self.overrides, **kwargs}
        for tag_name, value in overrides.items():
            tag = getattr(Tag, tag_name)
            if tag in dcm.keys():
                dcm[tag].value = value
            elif tag in dcm.file_meta:
                dcm.file_meta[tag].value = value
            else:
                elem = self.data_element(tag, value)
                dcm[tag] = elem

        pixels = self.pixels if pixels is None else pixels
        if pixels:
            arr = self.pixel_array(
                dcm.Rows,
                dcm.Columns,
                dcm.get("NumberOfFrames", 1),
                dcm.get("BitsStored", 16),
                dcm.get("BitsAllocated", 14),
                dcm.get("PhotometricInterpretation", "MONOCHROME1"),
                seed=self.seed,
            )
            dcm = set_pixels(dcm, arr, dcm.file_meta.TransferSyntaxUID)
        elif hasattr(dcm, "PixelData"):
            del dcm.PixelData

        # if requested, delete any tags not in the proto
        if not self.allow_nonproto_tags:
            for tag_name in overrides.keys():
                if tag_name not in self.dicom:
                    del dcm[tag_name]

        return dcm


FFDMFactory = partial(DicomFactory, Modality="MG", Laterality="L", ViewPosition="CC", NumberOfFrames=1)
TOMOFactory = partial(FFDMFactory, NumberOfFrames=3)
SynthFactory = partial(FFDMFactory, SeriesDescription="S-view")
UltrasoundFactory = partial(DicomFactory, Modality="US")
FACTORY_REGISTRY(FFDMFactory, name="ffdm")
FACTORY_REGISTRY(TOMOFactory, name="tomo")
FACTORY_REGISTRY(SynthFactory, name="synth")
FACTORY_REGISTRY(UltrasoundFactory, name="ultrasound")


class ConcatFactory(BaseFactory):
    def __init__(self, factories: Iterable[Union[str, DicomFactory, "ConcatFactory"]]):
        self.factories = [
            f if isinstance(f, (DicomFactory, ConcatFactory)) else FACTORY_REGISTRY.get(f)() for f in factories
        ]

    def __add__(self: C, other: C) -> C:
        result = copy(self)
        result.factories = self.factories + other.factories
        return result

    def __call__(self, seed: Optional[int] = None, pixels: Optional[bool] = None, **kwargs) -> List[FileDataset]:
        result: List[FileDataset] = []
        for factory in self.factories:
            _result = factory(seed=seed, pixels=pixels, **kwargs)
            if isinstance(_result, FileDataset):
                result.append(_result)
            else:
                result = result + _result
        return result


@FACTORY_REGISTRY(name="mammo-case")
class CompleteMammographyStudyFactory(ConcatFactory):
    def __init__(
        self,
        proto: Proto = DEFAULT_PROTO,
        seed: int = 42,
        allow_nonproto_tags: bool = True,
        implants: bool = False,
        spot_compression: bool = False,
        types: Iterable[str] = ("ffdm", "synth", "tomo"),
        lateralities: Iterable[str] = ("L", "R"),
        views: Iterable[str] = ("MLO", "CC"),
        **kwargs,
    ):
        IMPLANTS = (False, True) if implants else (False,)
        SPOT = (False, True) if spot_compression else (False,)

        lateralities = set(lateralities)
        types = set(types)
        views = set(views)

        factories: List[DicomFactory] = []
        iterator = product(lateralities, views, types, IMPLANTS, SPOT)
        for i, (laterality, view, mtype, implant, spot) in enumerate(iterator):
            meanings: List[str] = []
            if not implant and implants:
                meanings.append("implant displaced")
            if spot:
                meanings.append("spot compression")
            codes = DicomFactory.code_sequence(*meanings)
            overrides = {
                **kwargs,
                "ImageLaterality": laterality,
                "ViewPosition": view,
                "BreastImplantPresent": "YES" if implant else "NO",
                "ViewModifierCodeSequence": codes,
                "SOPInstanceUID": f"sop-{self.random_uid(seed=seed + i)}-{i}",
                "SeriesInstanceUID": f"series-{self.random_uid(seed=seed + i)}-{i}",
            }
            factory = cast(Type[DicomFactory], FACTORY_REGISTRY.get(mtype))(
                proto=proto,
                seed=seed,
                allow_nonproto_tags=allow_nonproto_tags,
                **overrides,
            )
            factories.append(factory)
        super().__init__(factories)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import re
from dataclasses import dataclass, field
from enum import IntEnum
from functools import cached_property, partial
from itertools import chain
from pathlib import Path
from typing import Any, Callable, Final, Generic, Iterable, List, Optional, Sequence, Sized, Tuple, TypeVar, Union

import numpy as np
import pydicom

from .dicom import Dicom, TransferSyntaxUIDs
from .metadata import MAX_FIELD_LENGTH
from .tags import Tag
from .types import PhotometricInterpretation


try:
    from colorama import Fore  # type: ignore
except Exception:
    Fore = None

T = TypeVar("T")


COLUMNS: Final = ("Tag", "Priority", "State", "Message")


def color_str(x: Any, color: str) -> str:
    if Fore is None:
        return str(x)
    elif not hasattr(Fore, color):
        raise AttributeError(f"Color {color} is not a valid Fore color")
    else:
        c = getattr(Fore, color)
        return f"{c}{x}{Fore.RESET}"


ValidationFunction = Callable[[T], Optional[str]]


def min_max_validator(min: float = float("-inf"), max: float = float("inf")) -> ValidationFunction:
    def func(val: Any) -> Optional[str]:
        try:
            if min <= val <= max:
                return None
            return "Value {val} outside interval [{min}, {max}]"
        except Exception:
            return f"Unhandled value {val}"

    return func


def int_validator(val: Any) -> Optional[str]:
    try:
        if int(val) == float(val):
            return None
    except Exception:
        pass
    return f"Value {val} is not an int"


def float_validator(val: Any) -> Optional[str]:
    try:
        float(val)
        return None
    except Exception:
        return f"Value {val} is not a float"


def age_validator(val: Any) -> Optional[str]:
    try:
        parsed_age = int(re.sub("[^0-9]", "", val))
    except Exception:
        return f"Unparsable age {val}"
    range_result = min_max_validator(20, 120)(parsed_age)
    return range_result


def transfer_syntax_validator(val: Any) -> Optional[str]:
    if val not in TransferSyntaxUIDs.keys():
        return f"Unknown TransferSyntaxUID {val}"
    elif "lossy" in (fullname := TransferSyntaxUIDs[val].lower()):
        return f"Detected lossy compression {fullname}"
    return None


def pixel_data_validator(val: Any) -> Optional[str]:
    if not len(val):
        return "Pixel data was empty"
    return None


def photometric_validator(val: Any) -> Optional[str]:
    pm = PhotometricInterpretation.from_str(val)
    if not pm:
        return f"Invalid value {val}"
    return None


def bits_validator(val: Any) -> Optional[str]:
    int_val = int_validator(val)
    if int_val is not None:
        return int_val
    val = int(val)
    if val <= 8:
        return f"Bit value {val} suggests low dynamic range"
    return None


class ValidationLevel(IntEnum):
    IGNORE = 0
    PRESENT = 1
    VALID = 2
    ABSENT = 3


class Priority(IntEnum):
    IGNORE = 0
    LOW = 1
    MEDIUM = 2
    HIGH = 3
    CRITICAl = 4

    def __repr__(self) -> str:
        return self.name

    def __str__(self) -> str:
        return self.name

    @classmethod
    def all_values(cls) -> Sequence["Priority"]:
        return list(cls)


@dataclass
class ValidationResult(Generic[T]):
    tag: Tag
    name: str
    level: ValidationLevel
    priority: Priority
    is_valid: bool
    value: Optional[T]
    msg: str

    @property
    def is_present(self) -> bool:
        if isinstance(self.value, (str, Sized)):
            return bool(len(self.value))
        return self.value is not None

    def __add__(self, other: "ValidationResult") -> "ValidationResult":
        return self if self.is_valid else other

    @property
    def result_level(self) -> ValidationLevel:
        if self.is_valid:
            return ValidationLevel.VALID
        elif self.is_present:
            return ValidationLevel.PRESENT
        else:
            return ValidationLevel.ABSENT

    @property
    def is_passing(self) -> bool:
        return self.result_level is not ValidationLevel.ABSENT and self.result_level >= self.level

    @property
    def _color(self) -> str:
        if self.is_valid:
            result_level = ValidationLevel.VALID
        elif self.is_present:
            result_level = ValidationLevel.PRESENT
        else:
            result_level = ValidationLevel.IGNORE

        if result_level >= self.level:
            return "GREEN"
        elif self.level == ValidationLevel.VALID:
            return "YELLOW"
        else:
            return "RED"

    @property
    def _passing_str(self) -> str:
        if self.is_passing:
            return "PASS"
        elif not self.is_present:
            return "MISSING"
        elif self.level == ValidationLevel.ABSENT:
            return "FAIL"
        else:
            return "INVALID"

    def to_line(self, fmt: str, color: bool = False) -> str:
        str_fn = partial(color_str, color=self._color) if color else str
        result = fmt.format(
            str_fn(str(self.tag)),
            str_fn(self.priority),
            str_fn(self._passing_str),
            str_fn(self.msg),
        )
        return result

    def header(self, fmt: str, color: bool = False) -> str:
        str_fn = partial(color_str, color="WHITE") if color else str
        result = fmt.format(*(str_fn(x) for x in COLUMNS))
        return result

    def column_widths(self, color: bool = False) -> List[int]:
        str_fn = partial(color_str, color=self._color) if color else str
        values = [
            str_fn(str(self.tag)),
            str_fn(self.priority),
            str_fn(self._passing_str),
            str_fn(self.msg),
        ]
        result = [max(len(v), len(color_str(c, "WHITE"))) for v, c in zip(values, COLUMNS)]
        return result


@dataclass
class ValidationTarget(Generic[T]):
    tag: Tag
    level: ValidationLevel
    priority: Priority = Priority.MEDIUM
    validator: ValidationFunction = field(default=lambda _: None, repr=False, hash=False)

    def __lt__(self, other: Union["ValidationTarget", Iterable["ValidationTarget"]]):
        if isinstance(other, ValidationTarget):
            return self.priority < other.priority
        elif isinstance(other, Sequence):
            return self.priority < max(x.priority for x in other)
        raise TypeError(other)

    def __gt__(self, other: Union["ValidationTarget", Iterable["ValidationTarget"]]):
        if isinstance(other, ValidationTarget):
            return self.priority > other.priority
        elif isinstance(other, Sequence):
            return self.priority > max(x.priority for x in other)
        raise TypeError(other)

    def __eq__(self, other: Union["ValidationTarget", Iterable["ValidationTarget"]]):
        if isinstance(other, ValidationTarget):
            return self.priority == other.priority
        elif isinstance(other, Sequence):
            return self.priority == max(x.priority for x in other)
        return False

    @cached_property
    def name(self) -> str:
        try:
            return pydicom.DataElement(self.tag, "US", None).name
        except Exception:
            return "Unnamed"

    def get_value(self, dcm: Dicom) -> Optional[T]:
        val = dcm.get(self.tag, None) or dcm.file_meta.get(self.tag, None)
        return val.value if val is not None else None

    def is_present(self, dcm: Dicom) -> bool:
        val = self.get_value(dcm)
        if isinstance(val, (str, Sized)):
            return bool(len(val))
        return val is not None

    def is_valid(self, dcm: Dicom) -> bool:
        result = self.is_present(dcm) and self.validator(self.get_value(dcm)) is None
        return result

    def validation_msg(self, dcm: Dicom) -> str:
        if not self.is_present(dcm):
            return "Field not present"
        result = self.validator(self.get_value(dcm))
        if result is not None:
            return result
        result = str(self.get_value(dcm))
        if len(result) <= MAX_FIELD_LENGTH:
            return result
        return "OK"

    def __call__(self, dcm: Dicom) -> ValidationResult:
        valid = self.is_valid(dcm)
        valid_msg = self.validation_msg(dcm)
        value = self.get_value(dcm)

        result = ValidationResult(
            self.tag,
            self.name,
            self.level,
            self.priority,
            is_valid=valid,
            value=value,
            msg=valid_msg,
        )
        return result


DEFAULT_TARGETS = [
    ValidationTarget(Tag.PatientIdentityRemoved, ValidationLevel.VALID, Priority.HIGH),
    ValidationTarget(Tag.StudyInstanceUID, ValidationLevel.VALID, Priority.HIGH),
    ValidationTarget(Tag.SeriesInstanceUID, ValidationLevel.VALID, Priority.HIGH),
    ValidationTarget(Tag.SOPInstanceUID, ValidationLevel.VALID, Priority.HIGH),
    ValidationTarget(Tag.TransferSyntaxUID, ValidationLevel.VALID, Priority.HIGH, validator=transfer_syntax_validator),
    ValidationTarget(Tag.PatientAge, ValidationLevel.VALID, validator=age_validator),
    ValidationTarget(Tag.Manufacturer, ValidationLevel.VALID),
    ValidationTarget(Tag.ManufacturerModelName, ValidationLevel.VALID),
    ValidationTarget(Tag.DeviceSerialNumber, ValidationLevel.VALID),
    ValidationTarget(Tag.Rows, ValidationLevel.VALID, validator=min_max_validator(min=1)),
    ValidationTarget(Tag.Columns, ValidationLevel.VALID, validator=min_max_validator(min=1)),
    ValidationTarget(Tag.PixelData, ValidationLevel.VALID, Priority.HIGH, validator=pixel_data_validator),
    (
        ValidationTarget(Tag.Laterality, ValidationLevel.VALID, Priority.HIGH),
        ValidationTarget(Tag.ImageLaterality, ValidationLevel.VALID, Priority.HIGH),
    ),
    ValidationTarget(Tag.ViewPosition, ValidationLevel.VALID, Priority.HIGH),
    ValidationTarget(Tag.PatientSex, ValidationLevel.VALID),
    ValidationTarget(Tag.EthnicGroup, ValidationLevel.VALID),
    ValidationTarget(Tag.BreastImplantPresent, ValidationLevel.VALID, Priority.HIGH),
    ValidationTarget(Tag.PhotometricInterpretation, ValidationLevel.VALID, Priority.HIGH, photometric_validator),
    ValidationTarget(Tag.PixelSpacing, ValidationLevel.VALID, Priority.LOW),
    ValidationTarget(Tag.BitsStored, ValidationLevel.VALID, Priority.LOW, bits_validator),
    ValidationTarget(Tag.PartialView, ValidationLevel.PRESENT, Priority.LOW),
    ValidationTarget(Tag.LossyImageCompression, ValidationLevel.VALID, Priority.LOW),
    ValidationTarget(Tag.Density, ValidationLevel.VALID, Priority.LOW),
    ValidationTarget(Tag.WindowCenter, ValidationLevel.IGNORE, Priority.LOW),
    ValidationTarget(Tag.WindowWidth, ValidationLevel.IGNORE, Priority.LOW),
]


class Validator:
    def __init__(self, targets: Sequence[Union[ValidationTarget, Tuple[ValidationTarget, ...]]]):
        self.targets = list(sorted(targets, reverse=True))

    def __add__(self, other: "Validator") -> "Validator":
        if not isinstance(other, Validator):
            raise TypeError(f"Expected Validator for other, found {type(other)}")
        return Validator(list(chain(self.targets, other.targets)))

    def validate_dicom_object(self, dcm: Dicom) -> Sequence[ValidationResult]:
        result: List[ValidationResult] = []
        for target in self.targets:
            if isinstance(target, Iterable):
                all_tag_results = [t(dcm) for t in target]
                tag_result = sum(all_tag_results, all_tag_results[0])
            else:
                tag_result = target(dcm)
            result.append(tag_result)
        return result

    def validate_dicom_file(self, path: Path) -> Sequence[ValidationResult]:
        if not path.is_file():
            raise FileNotFoundError(path)
        with pydicom.dcmread(path) as dcm:
            result = self.validate_dicom_object(dcm)
        return result

    @classmethod
    def default_validator(cls) -> "Validator":
        return cls(DEFAULT_TARGETS)

    @staticmethod
    def _format(
        results: Sequence[ValidationResult],
        delim: str = "\t",
        color: bool = False,
        padding: int = 2,
    ) -> str:
        lens = np.array([x.column_widths(color) for x in results])
        lens = lens.max(axis=0)
        lens += padding
        return delim.join([f"{{:<{x}}}" for x in lens])

    @classmethod
    def report_string(
        cls, results: Sequence[ValidationResult], failing_only: bool = False, color: bool = False
    ) -> None:
        passing = 0
        non_passing = {k: 0 for k in Priority.all_values()}
        fmt = cls._format(results, color=color)
        print(results[0].header(fmt, color=color))
        for result in results:
            if result.is_passing and failing_only:
                pass
            else:
                print(result.to_line(fmt, color=color))

            if result.is_passing:
                passing += 1
            else:
                non_passing[result.priority] += 1
        print("\n=== Summary === ")
        print(f"Passing: {passing}")
        print(f"Failing: {sum(v for v in non_passing.values())}")
        for k, v in non_passing.items():
            if v == 0:
                continue
            print(f"  {k.name}: {v}")

    @classmethod
    def all_passing(cls, results: Sequence[ValidationResult]) -> bool:
        return all(r.is_passing for r in results)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from signal import SIG_DFL, SIGINT, SIGPIPE, signal


# avoid problems w/ keyboard interrupt when pipeing
signal(SIGPIPE, SIG_DFL)
signal(SIGINT, SIG_DFL)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from copy import deepcopy
from pathlib import Path
from typing import Final

import numpy as np
import pydicom
import pytest
from pydicom.data import get_testdata_file
from pydicom.dataset import Dataset, FileMetaDataset
from pydicom.encaps import encapsulate
from pydicom.uid import (
    UID,
    DeflatedExplicitVRLittleEndian,
    ExplicitVRLittleEndian,
    ImplicitVRLittleEndian,
    RLELossless,
)


def get_rand_bytes(num: int) -> bytes:
    rand_bytes = np.random.default_rng().bytes(num)
    assert isinstance(rand_bytes, bytes)
    return rand_bytes


@pytest.fixture
def dicom_file():
    return get_testdata_file("CT_small.dcm")


@pytest.fixture
def dicom_object(dicom_file):
    with pydicom.dcmread(dicom_file) as dcm:
        yield dcm


@pytest.fixture
def dicom_file_j2k_uint16() -> str:
    filename = get_testdata_file("RG1_J2KR.dcm")
    assert isinstance(filename, str)
    return filename


@pytest.fixture
def dicom_file_j2k_int16() -> str:
    filename = get_testdata_file("JPEG2000.dcm")
    assert isinstance(filename, str)
    return filename


@pytest.fixture
def dicom_file_j2k_3d_uint16() -> str:
    filename = str(Path(__file__).parent / "jpeg2k_3d.dcm")
    assert isinstance(filename, str)
    return filename


@pytest.fixture
def dicom_file_jpl14() -> str:
    # JPEG lossless non-hierarchical (Process 14)
    filename = get_testdata_file("JPEG-LL.dcm")
    assert isinstance(filename, str)
    return filename


@pytest.fixture(params=["dicom_file_j2k_uint16", "dicom_file_j2k_int16", "dicom_file_j2k_3d_uint16"])
def dicom_file_j2k(request):
    return request.getfixturevalue(request.param)


@pytest.fixture
def dicom_file_3d(tmp_path, dicom_object_3d):
    path = Path(tmp_path, "CT_small_3D.dcm")
    dicom_object_3d(num_frames=32).save_as(path)
    return path


@pytest.fixture
def dicom_object_3d(dicom_object):
    series = "1.2.345"
    study = "2.3.456"
    dcm = deepcopy(dicom_object)
    source = Path(pydicom.data.get_testdata_files("*CT*")[0])  # type: ignore
    dcm.StudyInstanceUID = study
    dcm.SeriesInstanceUID = series
    dcm.SOPInstanceUID = series
    file_meta = FileMetaDataset()
    file_meta.MediaStorageSOPClassUID = UID("1.2.345")
    file_meta.MediaStorageSOPInstanceUID = UID("2.3.456")

    def func(num_frames, syntax=ExplicitVRLittleEndian):
        file_meta.TransferSyntaxUID = syntax
        old_data_len = len(dcm.PixelData)
        dcm.NumberOfFrames = num_frames

        if syntax.is_compressed:
            new_data = encapsulate([get_rand_bytes(old_data_len) for _ in range(num_frames)], has_bot=False)
            dcm.PixelData = new_data
            dcm.compress(syntax)
        else:
            dcm.PixelData = get_rand_bytes(old_data_len * num_frames)

        dcm.file_meta = file_meta
        return dcm

    return func


@pytest.fixture(
    params=[
        pytest.param(DeflatedExplicitVRLittleEndian, id="DeflatedExplicitVRLittleEndian"),
        pytest.param(ExplicitVRLittleEndian, id="ExplicitVRLittleEndian"),
        pytest.param(ImplicitVRLittleEndian, id="ImplicitVRLittleEndian"),
        pytest.param(RLELossless, id="RLELossless"),
    ]
)
def transfer_syntax(request):
    return request.param


NUM_DICOM_TEST_FILES: Final[int] = 3


@pytest.fixture(params=pydicom.data.get_testdata_files("*rgb*.dcm")[:NUM_DICOM_TEST_FILES])  # type: ignore
def test_dicom(request) -> Dataset:
    return pydicom.dcmread(request.param)



from concurrent.futures import Future, ThreadPoolExecutor
from dataclasses import dataclass
from enum import Enum
from typing import Any, Final, Iterator, List, NamedTuple, Optional, Tuple

import cv2
import numpy as np
from numpy import ndarray
from pydicom.uid import UID
from tqdm import tqdm

from .dicom import read_dicom_image
from .logging import logger
from .types import Dicom, DicomAttributeSequence


Color = Tuple[int, int, int]
Bbox = Tuple[int, int, int, int]
presentation_modality: Final[str] = "PR"


def kwargs2str(**kwargs: Any) -> str:
    key_val_str = ", ".join(f"{k.replace('_', ' ')}: {v}" for k, v in kwargs.items())
    return f"<{key_val_str}>"


class Form(Enum):
    CIRCLE = "CIRCLE"
    ELLIPSE = "ELLIPSE"
    POLYLINE = "POLYLINE"
    # POINT - defined by DICOM standard but not currently supported
    # INTERPOLATED - defined by DICOM standard but not currently supported

    def __init__(self, value) -> None:
        # Make sure CIRCLE is not assigned the name "CIRCL" or some other misspelling
        assert self.value == self.name == value, f"Enum name/value mismatch {self.value} != {self.name}"


@dataclass
class Reference:
    uid: UID
    frame: int  # The first frame is denoted as frame number 1

    @classmethod
    def from_graphic_annotation(cls, ann: Dicom) -> "Reference":
        return cls(UID(ann.ReferencedSOPInstanceUID), int(ann.get("ReferencedFrameNumber", 1)))

    def __repr__(self):
        return kwargs2str(SOPInstanceUID=self.uid, frame=self.frame)


@dataclass
class Annotation:
    """Store an annotation with corresponding DICOM filename"""

    def __init__(self, refs: List[Reference], data: List[float], form: Form) -> None:
        self.refs: List[Reference] = refs
        self.form: Form = Form(form)
        self.trace = dicom_trace_to_bbox(data, self.form)
        self.is_rectangle: bool = True  # TODO Add non-rectangular trace support

    @property
    def uids(self) -> List[UID]:
        return [ref.uid for ref in self.refs]

    def __repr__(self):
        return kwargs2str(references=self.refs, form=self.form, trace=self.trace, is_rectangle=self.is_rectangle)


@dataclass
class DicomImage:
    """Store DICOM image pixels with associated metadata"""

    pixels: ndarray
    uid: UID

    @classmethod
    def from_dicom(cls, dicom: Dicom, **kwargs) -> "DicomImage":
        pixels = to_rgb(read_dicom_image(dicom, **kwargs))
        return cls(pixels, dicom.SOPInstanceUID)

    @property
    def is_single_frame(self) -> bool:
        return self.pixels.shape[0] == 1

    def __repr__(self):
        return kwargs2str(SOPInstanceUID=self.uid, shape=self.pixels.shape)


class GraphicItem(NamedTuple):
    """Store info from DICOM graphic objects. More information about DICOM graphic objects can be found here:
    http://dicom.nema.org/medical/Dicom/2017c/output/chtml/part03/sect_C.10.5.html
    """

    data: List[float]
    form: Form

    def __add__(self, other: "GraphicItem") -> "GraphicItem":
        assert self.form == other.form == Form.POLYLINE, "Addition is only defined for POLYLINE items"
        return GraphicItem(self.data + other.data, self.form)


def to_collage(images: List[ndarray]) -> ndarray:
    num_images = len(images)

    assert num_images != 0, "There must be at least one image."
    assert all(len(i.shape) == 4 for i in images), "The images must have 4 dimensions."

    image_chns, _, max_image_rows, max_image_cols = np.array([i.shape for i in images]).max(axis=0)

    dtype = images[0].dtype
    collage_rows = 1 if num_images < 3 else 2
    collage_cols = int(num_images / collage_rows + 0.5)

    assert all(i.shape[1] == 3 for i in images)
    collage = np.zeros((image_chns, 3, collage_rows * max_image_rows, collage_cols * max_image_cols), dtype=dtype)

    for i, image in enumerate(images):
        row = int(i >= collage_cols)
        col = i % collage_cols
        start_row = row * max_image_rows
        start_col = col * max_image_cols
        image_chns, _, image_rows, image_cols = image.shape
        collage[:image_chns, :, start_row : start_row + image_rows, start_col : start_col + image_cols] = image

    return collage


def rint(x: float) -> int:
    return int(round(x))


def dicom_ellipse_to_bbox(data: List[float]) -> Bbox:
    assert len(data) != 4 * 3, "This function is not implemented to support 3D (x,y,z) coordinates"
    assert len(data) == 4 * 2, f"Invalid number of data points ({len(data)}) for a DICOM ellipse"
    major_x0, major_y0, major_x1, major_y1, minor_x0, minor_y0, minor_x1, minor_y1 = data
    xs = [rint(v) for v in [major_x0, major_x1, minor_x0, minor_x1]]
    ys = [rint(v) for v in [major_y0, major_y1, minor_y0, minor_y1]]
    x0, y0, x1, y1 = min(xs), min(ys), max(xs), max(ys)
    return x0, y0, x1, y1


def dicom_circle_to_bbox(data: List[float]) -> Bbox:
    assert len(data) != 2 * 3, "This function is not implemented to support 3D (x,y,z) coordinates"
    assert len(data) == 2 * 2, f"Invalid number of data points ({len(data)}) for a DICOM circle"
    x_center, y_center, x_border, y_border = data
    radius = np.sqrt((x_center - x_border) ** 2 + (y_center - y_border) ** 2)
    x0, y0, x1, y1 = [rint(v) for v in [x_center - radius, y_center - radius, x_center + radius, y_center + radius]]
    return x0, y0, x1, y1


def dicom_polylines_to_bbox(data: List[float]) -> Bbox:
    assert len(data) % 2 == 0, "This function is not implemented to support 3D (x,y,z) coordinates"
    xs = [rint(x) for x in data[::2]]
    ys = [rint(x) for x in data[1::2]]
    x0, y0, x1, y1 = min(xs), min(ys), max(xs), max(ys)
    return x0, y0, x1, y1


def dicom_trace_to_bbox(data: List[float], form: Form) -> Bbox:
    if form == Form.CIRCLE:
        return dicom_circle_to_bbox(data)
    elif form == Form.ELLIPSE:
        return dicom_ellipse_to_bbox(data)
    elif form == Form.POLYLINE:
        return dicom_polylines_to_bbox(data)
    else:
        raise Exception(f"Parsing is not supported for {form}")


def chw_to_hwc(image: ndarray) -> ndarray:
    """CxHxW -> HxWxC"""
    return np.rollaxis(image, 0, 3).copy()


def hwc_to_chw(image: ndarray) -> ndarray:
    """HxWxC -> CxHxW"""
    return np.rollaxis(image, 2, 0).copy()


def draw_rectangle(
    image: ndarray,
    coord0: Tuple[int, int],
    coord1: Tuple[int, int],
    color: Color = (0, 255, 0),
    thickness: int = 30,
) -> ndarray:
    # OpenCV can give weird errors without the following line
    image = cv2.UMat(image).get()  # type: ignore
    cv2.rectangle(image, coord0, coord1, color, thickness)
    return image


def dcms_to_images(dcms: List[Dicom], bar: bool = True, jobs: int = 4, **kwargs) -> Iterator[DicomImage]:
    """Some DICOMs may only contain annotations or other non-image information, so we want to
    pull out image data where applicable."""
    tqdm_bar = tqdm(desc="Loading images", total=len(dcms), disable=(not bar))

    def func(dcm: Dicom) -> Optional[DicomImage]:
        try:
            return DicomImage.from_dicom(dcm, **kwargs)
        except Exception as e:
            logger.info(e)

    def callback(f: Future):
        tqdm_bar.update(1)

    with ThreadPoolExecutor(jobs) as tp:
        futures: List[Future] = [tp.submit(func, dcm) for dcm in dcms]
        for f in futures:
            f.add_done_callback(callback)

        for f in futures:
            if result := f.result():
                yield result
    tqdm_bar.close()


def distance(a: List[float], b: List[float]) -> float:
    assert len(a) == len(b), "Expected lists of equal length for calculating distance"
    return sum((u - v) ** 2 for u, v in zip(a, b))


def are_contiguous_points(a: List[float], b: List[float]) -> bool:
    a_start = a[:2]
    a_stop = a[-2:]
    b_start = b[:2]
    return distance(a_start, a_stop) > distance(a_stop, b_start)


def are_contiguous_polylines(a: GraphicItem, b: GraphicItem) -> bool:
    return a.form == b.form == Form.POLYLINE and are_contiguous_points(a.data, b.data)


def group_polylines(graphic_items: List[GraphicItem]) -> Iterator[GraphicItem]:
    """Consecutive polyline traces may have been recorded separately when they were intended to be part of one single
    trace. Identify this situation and combine polylines accordingly."""
    graphic_items = graphic_items.copy()
    while graphic_items:
        item = graphic_items.pop(0)

        while graphic_items and are_contiguous_polylines(item, graphic_items[0]):
            item += graphic_items.pop(0)

        yield item


def gen_graphic_items(graphic_objects: DicomAttributeSequence) -> Iterator[GraphicItem]:
    """DICOM graphic objects contain info we don't really care about right now,
    so we distill the info interest into a generator of GraphicItems."""
    for graphic_object in graphic_objects:
        assert graphic_object.GraphicAnnotationUnits == "PIXEL"
        assert graphic_object.GraphicDimensions == 2
        yield GraphicItem(data=graphic_object.GraphicData, form=Form(graphic_object.GraphicType))


def dcm_to_annotations(dcm: Dicom, target_sop_uid: Optional[UID] = None) -> Iterator[Annotation]:
    """Search through a DICOM for graphic annotations, and only return annotations corresponding to
    a specific SOPInstanceUID if desired."""
    for graphic_annotation in dcm.get("GraphicAnnotationSequence", []):
        refs = [Reference.from_graphic_annotation(a) for a in graphic_annotation.ReferencedImageSequence]
        if target_sop_uid is None or target_sop_uid in [ref.uid for ref in refs]:
            # A TextObjectSequence may be present but no GraphicObjectSequence so ".get" is used
            graphic_items = list(gen_graphic_items(graphic_annotation.get("GraphicObjectSequence", [])))
            for graphic_item in group_polylines(graphic_items):
                yield Annotation(refs=refs, data=graphic_item.data, form=graphic_item.form)


def get_pr_reference_targets(dcm: Dicom) -> Optional[List[UID]]:
    targets = [uid for annotation in dcm_to_annotations(dcm) for uid in annotation.uids]
    return targets if targets else None


def dcms_to_annotations(dcms: List[Dicom], bar: bool = True) -> Iterator[Annotation]:
    dcms = [d for d in dcms if d.get("Modality", "") == presentation_modality]
    for dcm in tqdm(dcms, desc="Loading annotations", disable=(not bar), leave=bool(dcms)):
        try:
            yield from dcm_to_annotations(dcm)
        except Exception as e:
            logger.info(e)


def overlay_annotations_on_image(image: DicomImage, annotations: List[Annotation]) -> None:
    pixels = image.pixels
    assert pixels.shape[1] == 3, "Drawing assumes 3 color channels"
    assert pixels.dtype == np.uint8, "Drawing assumes 8-bits per color channel"

    for i, frame in enumerate(pixels):
        frame = chw_to_hwc(frame)

        for annotation in annotations:
            # The annotation is currently drawn on every frame, but we could selectively draw on only
            # the applicable frames by checking frames specified in annotation.refs
            assert annotation.is_rectangle, "Drawing for non-rectangular traces is not currently supported."
            x0, y0, x1, y1 = annotation.trace
            frame = draw_rectangle(frame, (x0, y0), (x1, y1))

        pixels[i] = hwc_to_chw(frame)


def overlay_annotations(images: List[DicomImage], annotations: List[Annotation]) -> None:
    for image in images:
        overlay_annotations_on_image(image, [ann for ann in annotations if image.uid in ann.uids])


def to_rgb(image: ndarray) -> ndarray:
    # 2D images will come as 1xHxW (treat as 1xHxW) and 3D images will come as 1xDxHxW (treat as DxHxW)
    assert image.shape[0] == 1, ""
    chns, rows, cols = image.shape[-3:]
    image = image.reshape(chns, rows, cols)

    # uint8 is sufficient for visualization and saves memory which is important for 3D images
    rgb = np.zeros((chns, 3, rows, cols), dtype=np.uint8)

    for i, channel in enumerate(image):
        for j in range(3):
            rgb[i, j, :, :] = to_8bit(channel)

    return rgb


def dcms_to_annotated_images(dcms: List[Dicom], **kwargs) -> List[DicomImage]:
    images = list(dcms_to_images(dcms, **kwargs))
    annotations = list(dcms_to_annotations(dcms))
    overlay_annotations(images, annotations)
    return images


def to_8bit(x: ndarray) -> ndarray:
    pixel_min, pixel_max = x.min(), x.max()
    delta = max(pixel_max - pixel_min, 1e-9)
    x = (x - pixel_min) / delta * 255
    return x.round().astype(np.uint8)
